/*version 1.4*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#define LOGBUFSIZE 32678

static int fd;
static const char *logfile_basename;
static int logfile_generations=10;
static long logfile_maxsize=1000000;

static void parse_arguments(int argc, char **argv) {
	int c;
	while((c=getopt(argc,argv,"vg:s:"))!=-1) {
		switch(c) {
			case 'v':
				printf("hplog - High Performance Logger v1.3\n");
				exit(EXIT_FAILURE);
			case 'g':
				logfile_generations = (int)strtol(optarg,0,0);
				if(logfile_generations<0)
					logfile_generations = 0;
				break;
			case 's':
				logfile_maxsize = strtol(optarg,0,0);
				if(logfile_maxsize<0)
					logfile_maxsize = 0;
				break;
		}
	}
	if(optind<argc)
		logfile_basename = argv[optind];
	else
		logfile_basename = "log";
}

static char logbuffer[LOGBUFSIZE];
static size_t bytes_in_logbuffer=0;
static void openlogfile() {
	fd = open(logfile_basename, O_CREAT|O_APPEND|O_WRONLY, 0644);
	if(fd<0) {
		perror("hplogd: could not open/create log file");
		exit(EXIT_FAILURE);
	}
}
static int flushlogfile() {
	if(bytes_in_logbuffer>0) {
		if(write(fd, logbuffer, bytes_in_logbuffer)!=(long)bytes_in_logbuffer)
			return -1;
		bytes_in_logbuffer = 0;
	}
	return 0;
}
static long writelogfile(const void *p, size_t bytes) {
	if(bytes_in_logbuffer+bytes>sizeof(logbuffer)) {
		if(flushlogfile()!=0)
			return -1;
	}
	memcpy(logbuffer+bytes_in_logbuffer, p, bytes);
	bytes_in_logbuffer += bytes;
	return bytes;
}
static void closelogfile() {
	flushlogfile();
	close(fd);
}
static void rotatelogfiles() {
	char tmp1[512];
	char tmp2[512];
	int i;
	sprintf(tmp1,"%s.%d",logfile_basename,logfile_generations);
	remove(tmp1);
	for(i=logfile_generations; i>0; i--) {
		sprintf(tmp1,"%s.%d",logfile_basename,i-1);
		sprintf(tmp2,"%s.%d",logfile_basename,i);
		rename(tmp1,tmp2);
	}
	rename(logfile_basename,tmp1);
}

int main(int argc, char **argv) {
	time_t last_sizecheck_time = 0;
	parse_arguments(argc,argv);
	openlogfile();
	fcntl(0, F_SETFL, O_NONBLOCK);
	signal(SIGPIPE,SIG_IGN);
	
	for(;;) {
		char buf[16384];
		long bytesread;
		int should_rotate;
		int should_closeopen;
		
		bytesread = read(0,buf,sizeof(buf));
		if(bytesread<=0) {
			fd_set fds;

			if(errno!=EAGAIN && errno!=EINTR) {
				/*Parent has closed connection*/
				close(0);
				closelogfile();
				return EXIT_SUCCESS;
			}
			flushlogfile();
			
			FD_ZERO(&fds);
			FD_SET(0,&fds);
			select(0+1, &fds,0,0, NULL);
			
			bytesread = read(0,buf,sizeof(buf));
			if(bytesread<=0) {
				close(0);
				closelogfile();
				return EXIT_SUCCESS;
			}
		}
		should_rotate=0;
		should_closeopen=0;
		if(last_sizecheck_time+1<=time(0)) {
			/*Time for checking if we should rotate log files*/
			last_sizecheck_time = time(0);
			flushlogfile();
			if(lseek(fd,0,2) >= logfile_maxsize) {
				should_rotate = 1;
				should_closeopen = 1;
			} else {
				struct stat disk_s,open_s;
				if(stat(argv[1],&disk_s)!=0 ||
				   fstat(fd,&open_s)!=0 ||
				   disk_s.st_dev!=open_s.st_dev ||
				   disk_s.st_ino!=open_s.st_ino)
				{
					/*Someone has renamed the file*/
				   	should_closeopen = 1;
				}
			}
		}
		if(should_rotate || should_closeopen) {
			char *p=(char*)memchr(buf,'\n',(size_t)bytesread);
			if(!p) p=buf;
			else p++;
			
			if(p!=buf)
				writelogfile(buf,p-buf);
			
			closelogfile();

			if(should_rotate)
				rotatelogfiles();

			openlogfile();
			if(p!=buf+bytesread)
				writelogfile(p,bytesread-(p-buf));
		} else {
			long rc = writelogfile(buf,(size_t)bytesread);
			if(rc!=bytesread) {
				if(errno==ENOSPC || errno==EFBIG) {
					closelogfile();
					rotatelogfiles();
					openlogfile();
					writelogfile(buf,bytesread);
				}
			}
		}
	}
}
